@@ -11,4 +11,4 @@ export GQUILT_ICON
 PYTHONPATH=$PYTHONPATH:$GQUILT_LIB_DIR
 export PYTHONPATH
 
-exec python $GQUILT_LIB_DIR/gquilt.pyc
+exec python $GQUILT_LIB_DIR/gquilt.py
