@@ -15,7 +15,17 @@
 ### along with this program; if not, write to the Free Software
 ### Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-import gtk, os, pango, gquilt_utils, time
+import gtk, os, pango, gquilt_utils, time, re
+
+def get_gtk_window(widget):
+    gtk_window = widget
+    while True:
+        temp = gtk_window.get_parent()
+        if temp:
+            gtk_window = temp
+        else:
+            break
+    return gtk_window
 
 def _read_fs_entity_name(flag, prompt, startdir, ok_button_label, sugg=""):
     dialog = gtk.FileChooserDialog(prompt, None, flag,
@@ -47,7 +57,11 @@ def read_directory_name(prompt="Directory ...", startdir=None, ok_button_label=g
 
 class _info_dialog(gtk.Dialog):
     def __init__(self, title=None, parent=None, flags=0, buttons=()):
-        gtk.Dialog.__init__(self, title, parent, flags, buttons)
+        try:
+            gtk.Dialog.__init__(self, title, parent, flags, buttons)
+        except:
+            parent=None
+            gtk.Dialog.__init__(self, title, parent, flags, buttons)
         if parent is None:
             self.set_position(gtk.WIN_POS_MOUSE)
         else:
@@ -86,6 +100,7 @@ class info_dialog(_info_dialog):
 FORCE = 1
 REFRESH_AND_RETRY = 2
 SKIP = 3
+EDIT = 4
 CANCEL = gtk.RESPONSE_CANCEL
 
 class info_force_dialog(_info_dialog):
@@ -101,6 +116,20 @@ class info_force_dialog(_info_dialog):
         self.destroy()
         return res == FORCE
 
+class info_force_or_edit_dialog(_info_dialog):
+    def __init__(self, msg=None, parent=None):
+        _info_dialog.__init__(self, "FYI", parent,
+                              gtk.DIALOG_DESTROY_WITH_PARENT,
+                              (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
+                               gtk.STOCK_EDIT, EDIT,
+                               "_Force", FORCE))
+        if msg != None:
+            self.set_message(msg)
+    def ask(self):
+        res = self.run()
+        self.destroy()
+        return res
+
 class info_force_or_skip_dialog(_info_dialog):
     def __init__(self, msg=None, parent=None):
         _info_dialog.__init__(self, "FYI", parent,
@@ -246,7 +275,7 @@ class update_save_dialog(gtk.Dialog):
     def __init__(self, title):
         gtk.Dialog.__init__(self, title, None,
             gtk.DIALOG_DESTROY_WITH_PARENT,
-            ("_Update", 1, "_Save", 2, "Save _As", 3,
+            (gtk.STOCK_REFRESH, 1, gtk.STOCK_SAVE, 2, gtk.STOCK_SAVE_AS, 3,
             gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
         self.connect("response", self._process_response)
         self.connect("close", self._process_close)
@@ -529,7 +558,7 @@ class dir_file_tree(file_tree):
         self._populate(self.root)
     def update(self, iter=None):
         if iter is None:
-            self._update_dir(".", None)
+            self._update_dir(os.curdir, None)
         else:
             path = self.store.fs_path(iter)
             if not os.path.exists(path):
@@ -646,68 +675,121 @@ class console(gtk.Frame):
         self.view.set_cursor_visible(False)
         self.view.set_editable(False)
         self.show_all()
+        self._finish_entry()
     def _append_to_console_tagged(self, msg, tag):
         self.text.insert_with_tags(self.text.get_end_iter(), msg, tag)
         self.view.scroll_to_mark(self.eobuf, 0.001)
-    def append_dat(self, suffix=""):
-        self._append_to_console_tagged(time.strftime("%Y-%m-%d %H:%M:%S") + suffix, self.dat_tag)
-    def echo_cmd(self, cmd):
-        self.append_dat(": ")
-        self._append_to_console_tagged(cmd, self.cmd_tag)
-    def append_stdout(self, msg):
-        self._append_to_console_tagged(msg, self.stdout_tag)
+    def _append_dat(self):
+        self._append_to_console_tagged(time.strftime("%Y-%m-%d %H:%M:%S") + ": ", self.dat_tag)
+    def _finish_entry(self):
+        self._append_to_console_tagged("%", self.dat_tag)        
+    def _log_entry(self, entry, tag):
+        entry = entry.rstrip()
+        if entry:
+            self._append_dat()
+            self._append_to_console_tagged(entry, tag)
+            self._append_newline()
+            self._finish_entry()
+    def log_entry(self, entry):
+        self._log_entry(entry, self.cmd_tag)
+    def log_entry_bold(self, entry):
+        self._log_entry(entry, self.dat_tag)
+    def _append_with_newline(self, entry, tag):
+        entry = entry.rstrip()
+        if entry:
+            self._append_to_console_tagged(entry, tag)
+            self._append_newline()
+    def _append_stdout(self, msg):
+        self._append_with_newline(msg, self.stdout_tag)
+    def _append_stderr(self, msg):
+        self._append_with_newline(msg, self.stderr_tag)
     def append_stderr(self, msg):
-        self._append_to_console_tagged(msg, self.stderr_tag)
-    def append_newline(self):
-        self._append_to_console_tagged("\n", self.stdout_tag)
-    def log_cmd(self, cmd, so, se):
-        self.echo_cmd(cmd)
-        if so is not None:
-            self.append_stdout(so)
-        if se is not None:
-            self.append_stderr(se)
-    def exec_cmd(self, cmd, args=""):
-        if args is not None and args != "":
-            cmd += " " + args
-        self.echo_cmd(cmd)
-        self.append_newline()
+        self._append_stderr(msg)
+        self._finish_entry()
+    def _append_newline(self):
+        self._append_to_console_tagged(os.linesep, self.stdout_tag)
+    def exec_cmd(self, cmd):
+        self._append_dat()
+        self._append_with_newline(cmd, self.cmd_tag)
+        while gtk.events_pending():
+            gtk.main_iteration()
         res, op, err = gquilt_utils.run_cmd(cmd)
-        self.append_stdout(op)
-        self.append_stderr(err)
+        self._append_stdout(op)
+        self._append_stderr(err)
+        self._finish_entry()
+        return (res, op, err)
+    def exec_cmd_with_alert(self, cmd):
+        res, op, err = self.exec_cmd(cmd)
         if res != 0:
             if len(err) > 0:
                 info_dialog(err).inform()
             else:
                 info_dialog(op).inform()
-    def exec_cmd_with_force(self, cmd, args="", force_option="-f"):
-        if args is None or args == "":
-            rcmd = cmd
-        else:
-            rcmd = cmd + " " + args
-            force_option += " " + args
-        self.echo_cmd(rcmd)
-        self.append_newline()
-        res, op, err = gquilt_utils.run_cmd(rcmd)
-        self.append_stdout(op)
-        self.append_stderr(err)
-        if res != 0:
-            if len(err) > 0:
-                emsg = err
-            else:
-                emsg = op
-            if info_force_dialog(emsg).ask():
-                self.exec_cmd(cmd, force_option)
 
-class scrolled_diff_text(gtk.Frame):
-    def __init__(self, frame_title=None):
-        gtk.Frame.__init__(self)
+STATES = [gtk.STATE_NORMAL, gtk.STATE_ACTIVE, gtk.STATE_PRELIGHT, gtk.STATE_PRELIGHT, gtk.STATE_INSENSITIVE]
+
+class tws_line_count_display(gtk.HBox):
+    def __init__(self):
+        gtk.HBox.__init__(self)
+        self.pack_start(gtk.Label("Added TWS lines:"), expand=False, fill=False)
+        self._entry = gtk.Entry()
+        self._entry.set_width_chars(1)
+        self._entry.set_text(str(0))
+        self._entry.set_editable(False)
+        self.pack_start(self._entry, expand=False, fill=False)
+        self.show_all()
+    def set_value(self, val):
+        sval = str(val)
+        self._entry.set_width_chars(len(sval))
+        self._entry.set_text(sval)
+        if val:
+            for state in STATES:
+                self._entry.modify_base(state, gtk.gdk.Color("#FF0000"))
+        else:
+            for state in STATES:
+                self._entry.modify_base(state, gtk.gdk.Color("#00FF00"))
+
+class scrolled_diff_text(gtk.VBox):
+    def __init__(self):
+        gtk.VBox.__init__(self)
         self.view = gtk.TextView()
         self.text = self.view.get_buffer()
+        self._action_group = gtk.ActionGroup("diff_tws_nav")
+        self._action_group.add_actions(
+            [
+                ("tws_nav_first", gtk.STOCK_GOTO_TOP, "_First", None,
+                 "Scroll to first line with added trailing white space",
+                 self._tws_nav_first_acb),
+                ("tws_nav_prev", gtk.STOCK_GO_UP, "_Prev", None,
+                 "Scroll to previous line with added trailing white space",
+                 self._tws_nav_prev_acb),
+                ("tws_nav_next", gtk.STOCK_GO_DOWN, "_Next", None,
+                 "Scroll to next line with added trailing white space",
+                 self._tws_nav_next_acb),
+                ("tws_nav_last", gtk.STOCK_GOTO_BOTTOM, "_Last", None,
+                 "Scroll to last line with added trailing white space",
+                 self._tws_nav_last_acb),
+            ])
+        self.tws_button_box = gtk.HBox()
+        for a_name in ["tws_nav_first", "tws_nav_prev", "tws_nav_next", "tws_nav_last"]:
+            button = gtk.Button(label="a_name")
+            action = self._action_group.get_action(a_name)
+            image = gtk.Image()
+            image.set_from_stock(action.get_property("stock_id"), gtk.ICON_SIZE_BUTTON)
+            button.set_image(image)
+            button.set_tooltip_text(action.get_property("tooltip"))
+            action.connect_proxy(button)
+            self.tws_button_box.pack_start(button)
+        self.tws_check = re.compile('^(\+.*\S)(\s+\n)$')
+        self.tws_list = []
+        self.tws_index = 0
+        self.tws_display = tws_line_count_display()
         self.index_tag = self.text.create_tag("INDEX", weight=pango.WEIGHT_BOLD, foreground="#0000AA", family="monospace")
         self.sep_tag = self.text.create_tag("SEP", weight=pango.WEIGHT_BOLD, foreground="#0000AA", family="monospace")
         self.minus_tag = self.text.create_tag("MINUS", foreground="#AA0000", family="monospace")
         self.lab_tag = self.text.create_tag("LAB", foreground="#AA0000", family="monospace")
         self.plus_tag = self.text.create_tag("PLUS", foreground="#006600", family="monospace")
+        self.added_tws_tag = self.text.create_tag("ADDED_TWS", background="#006600", family="monospace")
         self.star_tag = self.text.create_tag("STAR", foreground="#006600", family="monospace")
         self.rab_tag = self.text.create_tag("RAB", foreground="#006600", family="monospace")
         self.change_tag = self.text.create_tag("CHANGED", foreground="#AA6600", family="monospace")
@@ -719,8 +801,10 @@ class scrolled_diff_text(gtk.Frame):
         sw = gtk.ScrolledWindow()
         sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
         sw.add(self.view)
-        self.add(sw)
+        self.pack_start(sw)
+        self.pack_start(self.tws_button_box, expand=False, fill=True)
         self.show_all()
+        self.tws_button_box.hide_all()
     def _append_tagged_text(self, text, tag):
         self.text.insert_with_tags(self.text.get_end_iter(), text, tag)
     def _append_patch_line(self, line):
@@ -728,7 +812,13 @@ class scrolled_diff_text(gtk.Frame):
         if fc == " ":
             self._append_tagged_text(line, self.unchanged_tag)
         elif fc == "+":
-            self._append_tagged_text(line, self.plus_tag)
+            match = self.tws_check.match(line)
+            if match:
+                self._append_tagged_text(match.group(1), self.plus_tag)
+                self._append_tagged_text(match.group(2), self.added_tws_tag)
+                return len(match.group(1))
+            else:
+                self._append_tagged_text(line, self.plus_tag)
         elif fc == "-":
             self._append_tagged_text(line, self.minus_tag)
         elif fc == "!":
@@ -751,10 +841,25 @@ class scrolled_diff_text(gtk.Frame):
             self._append_tagged_text(line, self.rab_tag)
         else:
             self._append_tagged_text(line, self.index_tag)
+        return 0
     def set_contents(self, text):
+        old_count = len(self.tws_list)
         self.text.set_text("")
+        self.tws_list = []
+        line_no = 0
         for line in text.splitlines():
-            self._append_patch_line(line + "\n")
+            offset = self._append_patch_line(line + os.linesep)
+            if offset:
+                self.tws_list.append((line_no, offset))
+            line_no += 1
+        new_count = len(self.tws_list)
+        if new_count <= self.tws_index:
+            self.tws_index = new_count - 1
+        self.tws_display.set_value(new_count)
+        if old_count and not new_count:
+            self.tws_button_box.hide_all()
+        elif not old_count and new_count:
+            self.tws_button_box.show_all()
     def save_to_file(self, filename):
         try:
             file = open(filename, 'w')
@@ -764,4 +869,108 @@ class scrolled_diff_text(gtk.Frame):
         file.write(text)
         file.close()
         return True
+    def _scroll_to_index(self):
+        pos = self.tws_list[self.tws_index]
+        self.view.scroll_to_iter(self.text.get_iter_at_line_offset(pos[0], pos[1]), 0.01)
+    def _tws_nav_first_acb(self, action):
+        self.tws_index = 0
+        self._scroll_to_index()
+    def _tws_nav_prev_acb(self, action):
+        if self.tws_index:
+            self.tws_index -= 1
+        self._scroll_to_index()
+    def _tws_nav_next_acb(self, action):
+        self.tws_index += 1
+        if self.tws_index >= len(self.tws_list):
+            self.tws_index = len(self.tws_list) - 1
+        self._scroll_to_index()
+    def _tws_nav_last_acb(self, action):
+        self.tws_index = len(self.tws_list) - 1
+        self._scroll_to_index()
     
+_KEYVAL_UP_ARROW = gtk.gdk.keyval_from_name('Up')
+_KEYVAL_DOWN_ARROW = gtk.gdk.keyval_from_name('Down')
+
+class EntryWithHistory(gtk.Entry):
+    def __init__(self, max=0):
+        gtk.Entry.__init__(self, max)
+        self._history_list = []
+        self._history_index = 0
+        self._history_len = 0
+        self._saved_text = ''
+        self._key_press_cb_id = gtk.object("key_press_event", self._key_press_cb)
+    def _key_press_cb(self, widget, event):
+        if event.keyval in [_KEYVAL_UP_ARROW, _KEYVAL_DOWN_ARROW]:
+            if event.keyval == _KEYVAL_UP_ARROW:
+                if self._history_index < self._history_len:
+                    if self._history_index == 0:
+                        self._saved_text = self.get_text()
+                    self._history_index += 1
+                    self.set_text(self._history_list[-self._history_index])
+                    self.set_position(-1)
+            elif event.keyval == _KEYVAL_DOWN_ARROW:
+                if self._history_index > 0:
+                    self._history_index -= 1
+                    if self._history_index > 0:
+                        self.set_text(self._history_list[-self._history_index])
+                    else:
+                        self.set_text(self._saved_text)
+                    self.set_position(-1)
+            return True
+        else:
+            return False
+    def clear_text_to_history(self):
+        text = self.get_text().rstrip()
+        # don't save entries that start with white space
+        if text[0] in [' ', '\t']:
+            return
+        # no adjacent duplicate entries allowed
+        if (self._history_len == 0) or (text != self._history_list[-1]):
+            self._history_list.append(text)
+
+_KEYVAL_UP_ARROW = gtk.gdk.keyval_from_name('Up')
+_KEYVAL_DOWN_ARROW = gtk.gdk.keyval_from_name('Down')
+
+class EntryWithHistory(gtk.Entry):
+    def __init__(self, max=0):
+        gtk.Entry.__init__(self, max)
+        self._history_list = []
+        self._history_index = 0
+        self._history_len = 0
+        self._saved_text = ''
+        self._key_press_cb_id = self.connect("key_press_event", self._key_press_cb)
+    def _key_press_cb(self, widget, event):
+        if event.keyval in [_KEYVAL_UP_ARROW, _KEYVAL_DOWN_ARROW]:
+            if event.keyval == _KEYVAL_UP_ARROW:
+                if self._history_index < self._history_len:
+                    if self._history_index == 0:
+                        self._saved_text = self.get_text()
+                    self._history_index += 1
+                    self.set_text(self._history_list[-self._history_index])
+                    self.set_position(-1)
+            elif event.keyval == _KEYVAL_DOWN_ARROW:
+                if self._history_index > 0:
+                    self._history_index -= 1
+                    if self._history_index > 0:
+                        self.set_text(self._history_list[-self._history_index])
+                    else:
+                        self.set_text(self._saved_text)
+                    self.set_position(-1)
+            return True
+        else:
+            return False
+    def get_text_and_clear_to_history(self):
+        text = self.get_text().rstrip()
+        self.set_text("")
+        self._history_index = 0
+        # beware the empty command string
+        if not text:
+            return ""
+        # don't save entries that start with white space
+        if text[0] in [' ', '\t']:
+            return text.lstrip()
+        # no adjacent duplicate entries allowed
+        if (self._history_len == 0) or (text != self._history_list[-1]):
+            self._history_list.append(text)
+            self._history_len = len(self._history_list)
+        return text
