@@ -185,7 +185,7 @@ def get_patch_diff_lines(path):
 
 def append_lines_to_file(file, lines):
     for line in lines:
-        file.write(line + "\n")
+        file.write(line + os.linesep)
 
 def strip_trailing_ws(lines):
     n = len(lines)
