@@ -38,6 +38,48 @@ class quilt_commands(gquilt_utils.cmd_list):
                 return cmds
             index += 1
 
+# Find the value of a specified quilt environment variable.  These can be set
+# in as environment variables or in /etc/quilt.quiltrc, ~/.quiltrc or the file
+# specified by the QUILTRC environment variable (if they exist).  Only one of
+# the files will be read with precedence in the reverse of the order listed.
+# Any variable set in the file that is read will override the value for the
+# same variable that has been set as an environment variable.
+# So we need to do the same so we know what quilt will see.
+_default_env_var_value = { 'QUILT_PATCHES': 'patches', 'QUILT_SERIES': 'series' }
+
+def _get_quilt_env_var_value(vname, default_ok=False):
+    if default_ok:
+        try:
+            val = _default_env_var_value[vname]
+        except:
+            val = None
+    else:
+        val = None
+    try:
+        quiltrc = os.environ['QUILTRC']
+    except:
+        for fname in [os.path.expanduser("~/.quiltrc"), "/etc/quilt.quiltrc"]:
+            if os.path.exists(fname):
+                quiltrc = fname
+                break
+            else:
+                quiltrc = None
+    if quiltrc:
+        regex = re.compile("^\w*%s=(.*)$" % vname)
+        quiltrc_f = open(quiltrc, 'r')
+        for line in quiltrc_f.readlines():
+            m = regex.match(line.strip())
+            if m:
+                val = m.group(1)
+                break
+        quiltrc_f.close()
+    else:
+        try:
+            val = os.environ[vname]
+        except:
+            pass
+    return val
+
 def find_patchfns():
     binloc = gquilt_utils.which("quilt")
     if binloc is None:
@@ -54,23 +96,13 @@ patchfns = find_patchfns()
 if patchfns is None:
     qbsfe = None
 else:
-    qbsfe = 'bash -c ". ' + os.environ['GQUILT_LIB_DIR'] + '/qbsfe.sh" gquilt ' + patchfns
+    qbsfe = 'bash -c ". ' + os.path.join(os.environ['GQUILT_LIB_DIR'], "qbsfe" + os.extsep + "sh") + '" gquilt ' + patchfns
 
 def internal(cmd, input=None):
     if qbsfe is None:
         raise "quilt_specific_error", "Couldn't find patchfns"
     return gquilt_utils.run_cmd(" ".join([qbsfe, cmd]), input)
 
-# run a command and log the result to the provided console
-def _exec_console_cmd(console, cmd, error=gquilt_const.ERROR):
-    res, so, se = gquilt_utils.run_cmd(cmd)
-    if console is not None:
-        console.log_cmd(cmd + "\n", so, se)
-    if res != 0:
-        return (error, so, se)
-    else:
-        return (gquilt_const.OK, so, se)
-
 # Some useful private quilt functions (that shouldn't fail)
 def _patch_file_name(patch):
     res, so, se = internal(" ".join(["patch_file_name", patch]))
@@ -78,24 +110,28 @@ def _patch_file_name(patch):
         raise "quilt_specific_error", se
     return so[:-1]
 
+def _patch_basename(patchfile):
+    res, so, se = internal(" ".join(["basename", patchfile]))
+    if res != 0:
+        raise "quilt_specific_error", se
+    return so[:-1]
+
 def _patch_backup_dir(patch):
+    if _get_quilt_env_var_value("QUILT_PATCHES_PREFIX"):
+        patch = _patch_basename(patch)
     res, so, se = internal(" ".join(["backup_file_name", patch, "x"]))
     if res != 0:
         raise "quilt_specific_error", se
     return so[:-3]
 
 def _patch_backup_file(patch, filename):
+    if _get_quilt_env_var_value("QUILT_PATCHES_PREFIX"):
+        patch = _patch_basename(patch)
     res, so, se = internal(" ".join(["backup_file_name", patch, filename]))
     if res != 0:
         raise "quilt_specific_error", se
     return so[:-1]
 
-def _patch_files_may_have_changed(patch):
-    res, so, se = internal(" ".join(["files_may_have_changed", patch]))
-    if res != 0:
-        raise "quilt_specific_error", se
-    return res == 0
-
 def _convert_pop_res(res, so, se):
     if res != gquilt_const.OK:
         if re.search("needs to be refreshed first", so + se):
@@ -124,11 +160,17 @@ def _convert_import_res(res, so, se):
             res = gquilt_const.ERROR_FORCE
     return (res, so, se)
 
-def _quilt_version_ge(rmajor, rminor):
+def _quilt_version():
     res, so, se = gquilt_utils.run_cmd("quilt --version")
     if res != 0:
+        return None
+    return so
+
+def _quilt_version_ge(rmajor, rminor):
+    vstr = _quilt_version()
+    if not vstr:
         return False
-    vstrs = so.split(".")
+    vstrs = vstr.split(".")
     amajor = int(vstrs[0])
     if  amajor > rmajor:
         return True
@@ -138,18 +180,25 @@ def _quilt_version_ge(rmajor, rminor):
         return int(vstrs[1]) >= rminor
 
 def _patch_dir(dir=None):
-    try:
-        pd = os.environ['QUILT_PATCHES']
-    except:
-        pd = "patches"
+    pd = _get_quilt_env_var_value('QUILT_PATCHES', default_ok=True)
     if dir is None:
         dir = os.getcwd()
-    return dir + "/" + pd
+    return os.path.join(dir, pd)
+    
+def _series_file(dir=None):
+    sn = _get_quilt_env_var_value('QUILT_SERIES', default_ok=True)
+    return os.path.join(_patch_dir(dir), sn)
     
 # Now implement the tool interface for quilt
 class interface(gquilt_tool.interface):
     def __init__(self):
         gquilt_tool.interface.__init__(self, "quilt")
+    def is_available(self):
+        if _quilt_version() is None:
+            return False
+        return True
+    def requires(self):
+        return "\"quilt\" <http://savannah.nongnu.org/projects/quilt>"
     def new_playground(self, dir=None):
         patch_dir = _patch_dir(dir)
         if not os.path.exists(patch_dir):
@@ -157,23 +206,42 @@ class interface(gquilt_tool.interface):
                 os.makedirs(patch_dir)
             except os.error, value:
                return (gquilt_const.ERROR, "", value[1])
-        return gquilt_utils.run_cmd("touch " + os.path.join(patch_dir, "series"))
+        return gquilt_utils.run_cmd("touch " + _series_file(dir))
     def is_playground(self, dir=None):
-        patch_dir = _patch_dir(dir)
-        return os.path.exists(os.path.join(patch_dir, "series"))
+        root = self.get_playground_root(dir)
+        if not root:
+            return False
+        return os.path.exists(_series_file(root))
     def is_patch_applied(self, patch):
         res, so, se = internal("is_applied " + patch)
         return res == 0
     def patch_file_name(self, patch):
+        if _get_quilt_env_var_value("QUILT_PATCHES_PREFIX"):
+            return patch
         return _patch_file_name(patch)
     def top_patch(self):
         res, so, se = gquilt_utils.run_cmd("quilt top")
         if res == 0 or (se.strip() == "" and so.strip() == ""):
             return so.strip()
-        elif res == 512 and so.strip() == "":
+        elif (res == 512 or res == 2) and so.strip() == "":
             return ""
         else:
             raise "quilt_specific_error", se
+    def next_patch(self):
+        res, so, se = gquilt_utils.run_cmd("quilt next")
+        if res == 0 or (se.strip() == "" and so.strip() == ""):
+            return so.strip()
+        elif (res == 512 or res == 2) and so.strip() == "":
+            return ""
+        else:
+            raise "quilt_specific_error", se
+    def base_patch(self):
+        res, so, se = gquilt_utils.run_cmd("hg qapplied")
+        if res == 0 and so:
+            lines = so.splitlines(false)
+            return lines[0]
+        else:
+            return ""
     def last_patch_in_series(self):
         res, op, err = gquilt_utils.run_cmd("quilt series")
         if res != 0:
@@ -187,6 +255,20 @@ class interface(gquilt_tool.interface):
             pid = os.spawnlp(os.P_NOWAIT, "quilt", "quilt", "diff", difffld, files[0])
         else:
             pid = os.spawnlp(os.P_NOWAIT, "quilt", "quilt", "diff", difffld, "-P", patch, files[0])
+    def get_playground_root(self, dir=None):
+        pd = _get_quilt_env_var_value('QUILT_PATCHES', default_ok=True)
+        if not dir:
+            dir = os.getcwd()
+        root = dir
+        while True:
+            apd = os.path.join(root, pd)
+            if os.path.exists(apd) and os.path.isdir(apd):
+                return root
+            newroot = os.path.dirname(root)
+            if root == newroot:
+                break
+            root = newroot
+        return None
     def get_patch_status(self, patch):
         if not self.is_patch_applied(patch):
             return (gquilt_const.OK, gquilt_const.NOT_APPLIED, "")
@@ -195,7 +277,7 @@ class interface(gquilt_tool.interface):
         if (not os.path.exists(patch_fn)) or os.path.exists(patch_dirn + "~refresh"):
             needs_refresh = True
         else:
-            timestamp_fn = patch_dirn + "/.timestamp"
+            timestamp_fn = os.path.join(patch_dirn + ".timestamp")
             if not os.path.exists(timestamp_fn):
                 needs_refresh = True
             else:
@@ -233,10 +315,15 @@ class interface(gquilt_tool.interface):
                 #res, status, err = self.get_patch_status(pn)
                 #series.append((pn, status))
                 series.append((pn, gquilt_const.APPLIED))
-        try:
-            patch_dir = os.environ['QUILT_PATCHES']
-        except:
-            patch_dir = ""
+        # if the patches are in a non standard directory and patch names
+        # aren't being prefixed with that name then let the client know so
+        # that they can display the name of the patch directory in the header
+        if not _get_quilt_env_var_value("QUILT_PATCHES_PREFIX"):
+            patch_dir = _get_quilt_env_var_value('QUILT_PATCHES', default_ok=False)
+            if patch_dir == _default_env_var_value['QUILT_PATCHES']:
+                patch_dir = None
+        else:
+            patch_dir = None
         return (gquilt_const.OK, (patch_dir, series), "")
     def get_diff(self, filelist=[], patch=None):
         if patch is None:
@@ -248,7 +335,7 @@ class interface(gquilt_tool.interface):
             if not self.is_patch_applied(patch):
                 res, diff = gquilt_pfuns.get_patch_diff_lines(self.patch_file_name(patch))
                 if res:
-                    return (gquilt_const.OK, "\n".join(diff), [])
+                    return (gquilt_const.OK, os.linesep.join(diff), [])
             return (gquilt_const.ERROR, so, se)
         return (res, so, se)
     def get_combined_diff(self, start_patch=None, end_patch=None):
@@ -302,9 +389,7 @@ class interface(gquilt_tool.interface):
                 else:
                     cmd += "-p " + patch + " "
             cmd += newname
-            res, so, se = gquilt_utils.run_cmd(cmd)
-            if console is not None:
-                console.log_cmd(cmd + "\n", so, se)
+            res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
             if res != 0:
                 return (gquilt_const.ERROR, so, se)
             else:
@@ -339,13 +424,13 @@ class interface(gquilt_tool.interface):
         cmd = "quilt pop"
         if force:
             cmd += " -f"
-        res, so, se = _exec_console_cmd(console, cmd)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
         return _convert_pop_res(res, so, se)
     def do_push_patch(self, console, force=False):
         cmd = "quilt push"
         if force:
             cmd += " -f"
-        res, so, se = _exec_console_cmd(console, cmd)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
         return _convert_push_res(res, so, se)
     def do_pop_to_patch(self, console, patch=None):
         cmd = "quilt pop"
@@ -354,7 +439,7 @@ class interface(gquilt_tool.interface):
                 cmd += " -a"
             else:
                 cmd += " " + patch
-        res, so, se = _exec_console_cmd(console, cmd)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
         return _convert_pop_res(res, so, se)
     def do_push_to_patch(self, console, patch=None):
         cmd = "quilt push"
@@ -363,7 +448,7 @@ class interface(gquilt_tool.interface):
                 cmd += " -a"
             else:
                 cmd += " " + patch
-        res, so, se = _exec_console_cmd(console, cmd)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
         return _convert_push_res(res, so, se)
     def do_refresh_patch(self, console, patch=None, force=False):
         cmd = "quilt refresh"
@@ -371,15 +456,17 @@ class interface(gquilt_tool.interface):
             cmd += " -f"
         if patch is not None:
             cmd = " ".join([cmd, patch])
-        res, so, se = _exec_console_cmd(console, cmd)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd)
         if res != gquilt_const.OK:
             if re.search("Enforce refresh with -f", so + se):
                 res = gquilt_const.ERROR_FORCE
             else:
                 res = gquilt_const.ERROR
         return (res, so, se)
-    def do_create_new_patch(self, console, name):
-        return _exec_console_cmd(console, " ".join(["quilt", "new", name]))
+    def do_create_new_patch(self, console, name, force=False):
+        if force:
+            return (gquilt_const.ERROR, "", "\"quilt\" cannot force \"new\"")
+        return gquilt_tool.exec_console_cmd(console, " ".join(["quilt", "new", name]))
     def do_import_patch(self, console, filename, patchname=None, force=False):
         if patchname:
             cmd = "quilt import -P %s " % patchname
@@ -387,8 +474,10 @@ class interface(gquilt_tool.interface):
             cmd = "quilt import "
         if force:
             cmd += "-f -d n "
-        res, so, se = _exec_console_cmd(console, cmd + filename)
+        res, so, se = gquilt_tool.exec_console_cmd(console, cmd + filename)
         return _convert_import_res(res, so, se)
+    def do_finish_patch(sefl, console, patch):
+        return (gquilt_const.ERROR, "", '"quilt" does not have this feature')
     def do_merge_patch(self, console, patch):
         filename = _patch_file_name(patch)
         res, so, se = self.do_merge_patch_file(console, filename)
@@ -396,34 +485,45 @@ class interface(gquilt_tool.interface):
             self.do_delete_patch(console, patch)
         return (res, so, se)
     def do_merge_patch_file(self, console, filename):
-        return _exec_console_cmd(console, " < ".join(["quilt fold", filename]))
+        return gquilt_tool.exec_console_cmd(console, " < ".join(["quilt fold", filename]))
     def do_delete_patch(self, console, patch):
-        return _exec_console_cmd(console, " ".join(["quilt", "delete", patch]))
-    def do_add_files_to_patch(self, console, filelist, patch=None):
-        if patch == None:
-            cmd = "quilt add"
-        else:
-            if _quilt_version_ge(0, 43):
-                cmd = "quilt add -P " + patch
-            else:
-                cmd = "quilt add -p " + patch
+        return gquilt_tool.exec_console_cmd(console, " ".join(["quilt", "delete", patch]))
+    def do_add_files_to_patch(self, console, filelist):
+        cmd = "quilt add"
         # quilt will screw up semi silently if you try to add directories to a patch
         for f in filelist:
             if os.path.isdir(f):
                 return (gquilt_const.ERROR, "", f + ' is a directory.\n "quilt" does not handle adding directories to patches\n')
-        return _exec_console_cmd(console, " ".join([cmd, " ".join(filelist)]))
+        return gquilt_tool.exec_console_cmd(console, " ".join([cmd, " ".join(filelist)]))
     def do_remove_files_from_patch(self, console, filelist, patch=None):
-        if patch == None:
-            cmd = "quilt remove"
-        else:
-            if _quilt_version_ge(0, 43):
-                cmd = "quilt remove -P " + patch
+        if quilt_commands().has_cmd("remove"):
+            if patch == None:
+                cmd = "quilt remove"
             else:
-                cmd = "quilt remove -p " + patch
-        # quilt will screw up semi silently if you try to remove directories to a patch
-        for f in filelist:
-            if os.path.isdir(f):
-                return (gquilt_const.ERROR, "", f + ' is a directory.\n "quilt" does not handle removing directories from patches\n')
-        return _exec_console_cmd(console, " ".join([cmd, " ".join(filelist)]))
+                if _quilt_version_ge(0, 43):
+                    cmd = "quilt remove -P " + patch
+                else:
+                    cmd = "quilt remove -p " + patch
+            # quilt will screw up semi silently if you try to remove directories to a patch
+            for f in filelist:
+                if os.path.isdir(f):
+                    return (gquilt_const.ERROR, "", f + ' is a directory.\n "quilt" does not handle removing directories from patches\n')
+            return gquilt_tool.exec_console_cmd(console, " ".join([cmd, " ".join(filelist)]))
+        else:
+            return (gquilt_const.ERROR, "", '"quilt" does not handle removing files from patches\n')
+    def do_revert_files_in_patch(self, console, filelist, patch=None):
+        if quilt_commands().has_cmd("revert"):
+            if patch == None:
+                cmd = "quilt revert"
+            else:
+                cmd = "quilt revert -P " + patch
+            # quilt will screw up semi silently if you try to revert directories in a patch
+            for f in filelist:
+                if os.path.isdir(f):
+                    return (gquilt_const.ERROR, "", f + ' is a directory.'
+                            + os.linesep + '"quilt" does not handle reversion for directories in patches' + os.linesep)
+            return gquilt_tool.exec_console_cmd(console, " ".join([cmd, " ".join(filelist)]))
+        else:
+            return (gquilt_const.ERROR, "", '"quilt" does not handle reverting changes to in from patches\n')
     def do_exec_tool_cmd(self, console, cmd):
-        return _exec_console_cmd(console, " ".join(["quilt", cmd]))
+        return gquilt_tool.exec_console_cmd(console, " ".join(["quilt", cmd]))
